#include "emerging.hpp"

Node::Node(int l, int h, int s1, int s2, int c1, int c2) {
	this->l = l;
	this->h = h;
	this->s1 = s1;
	this->s2 = s2;
	this->c1 = c1;
	this->c2 = c2;
}

void Emerging::print(int* SA, unsigned char* t, int p, int l) {
	for (int i = SA[p]; (i < SA[p] + l) && ((char) t[i] > 1); i++) {
		cout << (char) t[i];
	}
  cout << endl;
}

// This method assumes the LCP-array is LCP[0..length] and
// LCP[length] = LCP[0] = -1. The SA is SA[0..length-1].
void Emerging::emerging(char* filename1, char* filename2, int format, float s, float g) {
  if (s <= 0.0 || s > 1.0 || g <  1.0) {
	cout << "Wrong parameters... exit!" << endl;
	return;
  }

  /**
   * FIRST PHASE: PREPROCESSING
   **/
  SAfor2DBnaive* SA4DB = new SAfor2DBnaive(filename1, filename2, format);
  int length = SA4DB->length + 1; // length of LCP (not counting last -1)
//  for (int i = 0; i < length; i++) cout << SA4DB->SA[i] << ","; cout << endl;
  LCPmanzini manzini(SA4DB->SA, SA4DB->t, length);
  int* LCP = manzini.constructDB();
//  for (int i = 0; i < length; i++) cout << LCP[i] << ","; cout << endl;
  RMQ_n_1_alstrup* rmq = new RMQ_n_1_alstrup(LCP, length+1);

  if (s*SA4DB->anz1 < 1) {
	cout << "Support must be at least 1/|db_1|... exit!" << endl;
	return;
  }

  /**
   * SECOND PHASE: LABELLING
   **/
  int* count1 = new int[length]; // counts number of LCAs for DB_1
  int* count2 = new int[length]; // ... and for DB_2
  int* last = new int[SA4DB->anz];
  for (int i = 0; i < length; i++) { count1[i] = 0; count2[i] = 0; }
  for (int i = 0; i < SA4DB->anz; i++) last[i] = 0;

  int entry; // # of entry in DB
  int pos;   // position of RMQ
  for (int i = 1; i < length; i++) {
	entry = SA4DB->index(SA4DB->SA[i]);
	if (last[entry] != 0) {
	  pos = rmq->query(last[entry]+1,i);
	  if (entry < SA4DB->anz1) count1[pos]++; else count2[pos]++;
	}
	last[entry] = i;
  }

  /**
   * THIRD PHASE: EXTRACTION
   **/
  stack<Node*> S;             // stack for DF-traversal (keeps nodes)
  Node *node, *node2;         // variables needed for the DF-traversal
  int left;                   // first index of string in SA
  int olds1, olds2;           // s-values of last completely visited subtree
  int oldc1, oldc2;           // c-values of last completely visited subtree
  int h_parent;               // stores height of parent node (for output)
  float support;
  float growthrate;

  S.push(new Node(0, -1, -1, -1)); // virtual root (acts as stopper)
  for (int k = 1; k <= length; k++) { // start at 1 to skip the "$"!
	node = S.top(); // go to next leaf
	left = k-1;
	olds1 = 0; olds2 = 0; oldc1 = 0; oldc2 = 0;
	while (node->h > LCP[k]) {
	  S.pop();              // the popped node is now in 'node'

      // propagate counters upwards:
	  node2 = S.top();
	  if (node2->h >= LCP[k]) { // otherwise top-node not parent node
		node2->s1 += node->s1;
		node2->s2 += node->s2;
		node2->c1 += node->c1;
		node2->c2 += node->c2;
		h_parent = node2->h; // for outputting all patterns
	  }
	  else h_parent = LCP[k];

	  support = ((float) (node->s1 - node->c1)) / ((float) SA4DB->anz1);
	  growthrate = support / (((float) (node->s2 - node->c2))/ ((float) SA4DB->anz2));

      if (support >= s && growthrate >= g &&
		SA4DB->t[SA4DB->SA[node->l]+node2->h] > '\1' &&
		SA4DB->t[SA4DB->SA[node->l]+LCP[k]] > '\1') { // This is an Emerging Substring!
		// calculate true length of string s to be printed
		// (time for loop is O(|s|))
		for (int h = 0; h < node->h; h++) {
		  if (SA4DB->t[SA4DB->SA[node->l]+h] <= '\1') {
			node->h = h; break;
		  }
		}
		for (int h = h_parent + 1; h <= node->h; h++) {
			print(SA4DB->SA, SA4DB->t, node->l, h);
		}
 	  }
	  left = node->l;   // update left border
	  olds1 = node->s1; olds2 = node->s2;
	  oldc1 = node->c1; oldc2 = node->c2;
	  delete node;      // free space
	  node = S.top();   // next node on stack
	}
	if (node->h < LCP[k]) { // push lca (every inner node IS lca of consec. leaves!)
	  S.push(new Node(left, LCP[k], olds1, olds2, oldc1, oldc2));
	}

	// accumulate correction factors from arrays count1[], count2[]:
	node2 = S.top();
	node2->c1 += count1[k];
	node2->c2 += count2[k];

    // push leaf:
	if (k < length) {
	  entry = SA4DB->index(SA4DB->SA[k]); // check if current string is from DB1 or DB2
	  pos = entry < SA4DB->anz1 ? 1 : 0; 
	  S.push(new Node(k, length - SA4DB->SA[k], pos, 1-pos));
	}
  }
  delete S.top();
  delete[] LCP;
  delete[] count1;
  delete[] count2;
  delete[] last;
  delete rmq;
}
